#include "eureka.h"

#define SLAVE "node"


FILE *logfile, *infile;
char start_time[27], finish_time[27];
short sample=0;
int cluster_size, num_clusters, mytid, nhost, narch, goal_cluster, goal_tid;
int goal[16] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15};
int threshold_status[MAX_PROCESSORS][3], processor_level[MAX_PROCESSORS];
float imbalance = -1.0;
struct pvmhostinfo *hostp[MAX_PROCESSORS];
time_t t1, t2, t3;
List *BFS = NULL;

main(argc, argv)
   int argc;
   char *argv[];
{
   int i;

   mytid = pvm_mytid();                                      /* enroll in pvm */
   pvm_catchout(stdout);

   pvm_config(&nhost, &narch, hostp);        /* set number of slave processes */
   numnodes = nhost;

   InitLog();
   GetEnvironVars();

   if ((infile = fopen(datafile, "r")) == NULL) {    /* set up the input file */
      fprintf(stderr, "Unable to open the datafile %s\n", datafile);
      exit(1);
   }

   if (sample) {
      numnodes = 1;
      distribution_type = KUMAR_RAO;
      ordering = NONE;
      num_clusters = 1;
      cluster_size = 1;
      optimal = TRUE;
      fprintf(logfile, "SAMPLE\n");
      Search();
      GetEnvironVars();
      sample = 0;
   }

   OutputLog();
   Search();

   pvm_exit();

   fclose(logfile);
}


Search()
{
   int i, dest, cluster_num, processor_status[MAX_PROCESSORS];
   int status = 1, prev_goal_level, levelFlag, doneFlag, tid, length;

   for (i=1; i<numnodes; i++)            /* initially all processors are idle */
      processor_status[i] = 0;

   InitializeNodes();                            /* initialize the processors */
   InitializeData();
   SendData();

   status = NO_MESSAGE;
   levelFlag = FALSE;                 /* keep track of lower threshold values */
   lowest_heuristic = prev_goal_level = goal_level = 999;

   while (1) {                                                 /* search loop */
      rc = pvm_recv(-1, -1);                 /* receive status from processor */
      pvm_bufinfo(rc, &length, &status, &tid);
      dest = find_processor(tid);        /* identify processor sending status */
      cluster_num = dest / cluster_size;   /* identify cluster sending status */
      if (status == MOVING_ON) { /* processor is continuing to next threshold */
         pvm_upkint(&processor_level[dest], 1, 1);
         fprintf(stderr, "processor %d is moving on to threshold %d\n",
	    dest, processor_level[dest]);  fflush(stderr);
         if (levelFlag) {
	    doneFlag = TRUE;   /* see if any processors have lower thresholds */
	    for (i=0; i<numnodes; i++)
	       if (processor_level[i] < goal_level) {
	          doneFlag = FALSE;
		  break;
	       }
	    if (doneFlag)                               /* search is finished */
	       break;
	 }
      }
      else if (status == IDLE)
         processor_status[dest] = 1;
      else if (status == BUSY)
         processor_status[dest] = 0;
      else if (status == EUREKA) {                    /* processor found goal */
         pvm_upkint(&goal_level, 1, 1);
	 if (goal_level < prev_goal_level) {
	    pvm_initsend(PvmDataDefault);          /* send updated goal level */
	    pvm_pkint(&goal_level, 1, 1);
	    pvm_mcast(tids, numnodes, GOAL_LEVEL_COMING);
	    prev_goal_level = goal_level;
	 }
	 goal_tid = tid;                        /* possible winning processor */
	 threshold_status[cluster_num][1] = TRUE;       /* cluster found goal */
	 threshold_status[cluster_num][2] = dest;        /* winning processor */
	 if ((optimal == TRUE) && (load_balancing == ASYNC)) {
	    levelFlag = doneFlag = TRUE;
	    for (i=0; i<numnodes; i++)
	          if (processor_level[i] < goal_level) {
		     doneFlag = FALSE;
		     break;
	    }
	    if (doneFlag)
	       break;
	 }
      }

                                    /* update completed threshold information */
      if (load_balancing != ASYNC) {
         if (update_threshold(processor_status, cluster_num))
	    break;
      }

      if (least_cost_goal())                                      /* finished */
         break;
   }

   if (sample == 0)
      GetStats();
   else return;

   for (i=0; i<numnodes; i++)
      pvm_kill(tids[i]);
}


InitializeData()
{
   int i;

   for (i=0; i<16; i++)               /* read 15 puzzle initial configuration */
      fscanf(infile, "%d", &puzzle[i]);

   for (i=0; i<MAXOPS; i++)
      oporder[i] = initial_oporder[i];

   if (distribution_type == DTS) {
      BFS = MakeEmptyList(BFS);
      bfs(BFS, cluster_size, puzzle, goal, &threshold);
   }
   else if (initial_threshold == -1)
      threshold = state_cost(puzzle, goal);            /* threshold = h(root) */
   else threshold = initial_threshold;    /* threshold is user-supplied value */
   fprintf(logfile, "initial_threshold is %d\n", initial_threshold);

   for (i=0; i<numnodes; i++)   /* store the threshold used by each processor */
      processor_level[i] = threshold;

   total_nodes = prev_nodes = 0;
}


/* Collect information about completed search. */
GetStats()
{
   int i, dest, level, path[MAX_DEPTH];
   long hours, minutes, seconds;

   if (goal_cluster != -1)                            /* Get goal information */
      dest = tids[threshold_status[goal_cluster][2]];   /* least cost cluster */
   else dest = goal_tid;                                    /* first solution */
   pvm_initsend(PvmDataDefault);
   pvm_send(dest, SEND_GOAL_INFO);

   pvm_recv(dest, GOAL_INFO);
   pvm_upkint(&level, 1, 1);      /* receive level where goal was found */
   pvm_upkint(oporder, MAXOPS, 1);     /* receive updated oporder */
                       /* receive number of nodes generated on that processor */
   pvm_upkint(&i, 1, 1);
   pvm_upkint(path, level, 1);    /* receive path to the goal */
   pvm_upklong(&t2, 1, 1);        /* receive finish time on that processor */

   total_nodes += i;
   strncpy(finish_time, ctime(&t2) + 11, 8);         /* get timing statistics */
   finish_time[8] = '\0';
   t3 = t2 - t1;
   hours = t3 / 3600;
   minutes = (t3 - (hours * 3600)) / 60;
   seconds = t3 - (hours * 3600) - (minutes * 60);

   fprintf(logfile, "Puzzle:              ");       /* save stats in log file */
   for (i=0; i<16; i++)
      fprintf(logfile, "%3d", puzzle[i]);
   fprintf(logfile, "\n");
   fprintf(logfile, "Total nodes:    %8d\n", total_nodes);
   fprintf(logfile, "Evaluation order:    ");
   for (i=0; i<MAXOPS; i++)
      fprintf(logfile, "%d ", oporder[i]);
   fprintf(logfile, "\n");
   fprintf(logfile, "Goal found on processor:  %5d\n", dest);
   fprintf(logfile, "Goal found on level:  %4d\n\n", level);
   fprintf(logfile, "Start time:   %s\n", start_time);
   fprintf(logfile, "Finish time:  %s\n", finish_time);
   fprintf(logfile, "              -----------\n");
   fprintf(logfile, "Elapsed:      %02d:%02d:%02d\n", hours, minutes, seconds);
   fprintf(logfile, "Seconds:      %8d\n", t3);
   fflush(logfile);
}


/* Send initial data to each processor. */
SendData()
{
   int i, j, cluster_num;

   pvm_initsend(PvmDataDefault);    /* send pvm nodes information */
   pvm_pkint(&numnodes, 1, 1);
   pvm_pkint(tids, numnodes, 1);
   pvm_mcast(tids, numnodes, NODE_INFO);

   pvm_initsend(PvmDataDefault);    /* send goal specification */
   pvm_pkint(goal, 16, 1);
   pvm_mcast(tids, numnodes, GOAL_COMING);

   pvm_initsend(PvmDataDefault);    /* send operator order */
   pvm_pkint(oporder, MAXOPS, 1);
   pvm_mcast(tids, numnodes, OPORDER_COMING);

   pvm_initsend(PvmDataDefault);    /* send load balancing information */
   pvm_pkint(&load_balancing, 1, 1);
   pvm_mcast(tids, numnodes, LOAD_BAL_SWITCH_COMING);

   pvm_initsend(PvmDataDefault);    /* send ordering information */
   pvm_pkint(&ordering, 1, 1);
   pvm_mcast(tids, numnodes, ORDERING_SWITCH_COMING);

   pvm_initsend(PvmDataDefault);    /* send alpha */
   pvm_pkfloat(&alpha, 1, 1);
   pvm_mcast(tids, numnodes, ALPHA_COMING);

   pvm_initsend(PvmDataDefault);    /* send giveaway information */
   pvm_pkint(&giveaway, 1, 1);
   pvm_mcast(tids, numnodes, GIVEAWAY_SWITCH_COMING);

   pvm_initsend(PvmDataDefault);    /* send anticipatory information */
   pvm_pkint(&anticipatory, 1, 1);
   pvm_mcast(tids, numnodes, ANTICIPATORY_SWITCH_COMING);

   for (i=0; i<num_clusters; i++)   /* distribute search nodes to processors */
      InitialDistribution(BFS, distribution_type, puzzle,
                          i*cluster_size, cluster_size);

   for (i=0; i<num_clusters; i++) { /* send initial threshold to all clusters */
      fprintf(logfile, "Starting threshold %d to cluster %d\n", threshold, i);
      fprintf(stderr, "Starting threshold %d to cluster %d\n", threshold, i);
      fflush(stderr);
      for (j=0; j<cluster_size; j++) {
         pvm_initsend(PvmDataDefault);
         pvm_pkint(&threshold, 1, 1);
	 pvm_send(tids[i*cluster_size + j], THRESHOLD_COMING);
      }
      threshold_status[i][0] = threshold;       /* threshold for this cluster */
      threshold_status[i][1] = FALSE;                 /* cluster not finished */
      threshold += THRESHOLD_INCREMENT;
   }
   threshold -= THRESHOLD_INCREMENT;

   t1 = time((time_t *) 0);                               /* store start time */
   strncpy(start_time, ctime (&t1) + 11, 8);
   start_time[8] = '\0';

   pvm_initsend(PvmDataDefault);  /* tell processors to start */
   pvm_mcast(tids, numnodes, GOFORIT);
}


int Test(testtype, temp)
   int testtype;
   char *temp;
{
   char Data[5][50], consultfile[60];

   fprintf(stderr, "In Test, consulting file %s\n", cfile);
   if (sample == 0)
      sample = 1;

   if (imbalance == -1.0)
      return(0);

   sprintf(Data[0], "%d", (int) 1);                  /* problem number */
   sprintf(Data[1], "%f", bf);                       /* branching factor */
   sprintf(Data[2], "%f", avg_herror);               /* heuristic error */
   sprintf(Data[3], "%f", imbalance);                /* imbalance */
   sprintf(Data[4], "%f", hbf);                      /* hbf */
   fprintf(stderr, "Consulting with testtype %d %s %s %s %s\n",
      testtype, Data[1], Data[2], Data[3], Data[4]); fflush(stderr);

   switch(testtype) {
      case 0: {       /* distribtype */
         sprintf(consultfile, "%s/distrib", cfile);
         consult(Data, consultfile, temp);
         break;
      }
      case 1: {       /* loadbal */
         sprintf(consultfile, "%s/loadbal", cfile);
         consult(Data, consultfile, temp);
         break;
      }
      case 2: {       /* ordering */
         sprintf(consultfile, "%s/ordering", cfile);
         consult(Data, consultfile, temp);
         break;
      }
      case 3: {       /* alpha */
         sprintf(consultfile, "%s/alpha", cfile);
         consult(Data, consultfile, temp);
	 break;
      }
      case 4: {       /* giveaway */
         sprintf(consultfile, "%s/giveaway", cfile);
         consult(Data, consultfile, temp);
	 break;
      }
      case 5: {       /* numclusters */
         sprintf(consultfile, "%s/clusters", cfile);
         consult(Data, consultfile, temp);
         break;
      }
      case 6: {       /* anticipatory */
         sprintf(consultfile, "%s/anticip", cfile);
         consult(Data, consultfile, temp);
         break;
      }
      default:  break;
   }

   fprintf(logfile, "Consulted value of %d is %s\n", testtype, temp);
   fprintf(stderr, "Consulted value of %d is %s\n", testtype, temp);
   return(1);
}


/* Get the values of appropriate environment variables. */
GetEnvironVars()
{
   char temp[50];
   int i;

   strcpy(datafile, getenv("datafile"));   /* problem data file */
   if (datafile == NULL) {
      fprintf(logfile, "ERROR:  Datafile must be a valid input file \n");
      fprintf(stderr, "ERROR - See log file for details.\n");
      exit(1);
   }

   strcpy(temp, getenv("numnodes"));      /* number of processors */
   if (temp == NULL) {
      fprintf(logfile, "ERROR:  The environment var numnodes must be set to\n");
      fprintf(logfile, "        an integer value.\n");
      fprintf(stderr, "ERROR - See log file for details.\n");
      exit(1);
   }
   numnodes = atoi(temp);

   strcpy(temp, getenv("distribtype"));   /* task distribution technique */
   if ((temp == NULL) ||
       ((strcmp(temp, "KUMAR") != 0) && (strcmp(temp, "DTS") != 0))) {
      fprintf(logfile,
         "ERROR:  Distribtype must = KUMAR, DTS, or CDISTRIB\n");
      fprintf(stderr, "ERROR - See log file for details.\n");
      exit(1);
   }
   if (strcmp(temp, "CDISTRIB") == 0) {
      if (Test(0, temp) == 0)
         strcpy(temp, "KUMAR");
   }
   if (!strcmp(temp, "KUMAR"))
      distribution_type = KUMAR_RAO;
   else distribution_type = DTS;

   strcpy(cfile, getenv("consultdir"));   /* C4.5 rules directory */

   strcpy(temp, getenv("loadbal"));       /* load balancing technique */
   if ((temp == NULL) ||
       ((strcmp(temp, "NEIGHBOR") != 0) && (strcmp(temp, "RANDOM") != 0) &&
        (strcmp(temp, "ASYNC") != 0) && (strcmp(temp, "OFF") != 0))) {
      fprintf(logfile,
         "ERROR:  Loadbal must = NEIGHBOR, RANDOM, ASYNC, OFF, or CLOADBAL\n");
      fprintf(stderr, "ERROR - See log file for details.\n");
      exit(1);
   }
   if (strcmp(temp, "CLOADBAL") == 0) {
      if (Test(1, temp) == 0)
         strcpy(temp, "OFF");
   }
   if (!strcmp(temp, "NEIGHBOR"))
      load_balancing = NEIGHBOR;
   else if (!strcmp(temp, "RANDOM"))
      load_balancing = RANDOM;
   else if (!strcmp(temp, "ASYNC"))
      load_balancing = ASYNC;
   else load_balancing = OFF;

   strcpy(temp, getenv("ordering"));       /* tree ordering technique */
   if ((temp == NULL) ||
       ((strcmp(temp, "TOIDA") != 0) &&
        (strcmp(temp, "LOCAL") != 0) && (strcmp(temp, "NONE") != 0))) {
      fprintf(logfile, "ERROR:  Ordering must = LOCAL, TOIDA, NONE, or CORDER\n");
      fprintf(stderr, "ERROR - See log file for details.\n");
      exit(1);
   }
   if (strcmp(temp, "CORDER") == 0) {
      if (Test(2, temp) == 0)
         strcpy(temp, "NONE");
   }
   if (!strcmp(temp, "TOIDA"))
      ordering = TOIDA;
   else if (!strcmp(temp, "LOCAL"))
      ordering = LOCAL;
   else ordering = NONE;

   strcpy(temp, getenv("order"));         /* initial operator order */
   if (temp == NULL) {
      fprintf(logfile, "ERROR:  The environment var order was not set -\n");
      fprintf(logfile, "        default is used.\n");
      fprintf(stderr, "WARNING - See log file for details.\n");
      for (i=0; i<MAXOPS; i++)
         initial_oporder[i] = i;
   }
   else {
      for (i = 0; i<strlen(temp); i++)
         initial_oporder[i] = temp[i] - '0';
   }

   strcpy(temp, getenv("threshold"));
   if (temp == NULL) {
      fprintf(logfile,
         "ERROR:  Threshold must be defined as -1 or a positive integer\n");
      fprintf(stderr, "ERROR - See log file for details.\n");
      exit(1);
   }
   initial_threshold = atoi(temp);

   strcpy(temp, getenv("alpha"));     /* percentage giveaway */
   if (temp == NULL) {
      fprintf(logfile,
         "ERROR:  Alpha must be defined as a float\n");
      fprintf(stderr, "ERROR - see log file for details.\n");
      exit(1);
   }
   if (strcmp(temp, "CALPHA") == 0) {
      if (Test(3, temp) == 0)
         strcpy(temp, "0.3");
   }
   alpha = (float) atof(temp);

   strcpy(temp, getenv("giveaway"));  /* give from front or back of OPEN list */
   if (strcmp(temp, "CGIVEAWAY") == 0) {
      if (Test(4, temp) == 0)
         strcpy(temp, "BOTTOM");
   }
   if (!strcmp(temp, "BOTTOM"))
      giveaway = BOTTOM;
   else giveaway = TOP;

   strcpy(temp, getenv("numclusters"));    /* number of processor clusters */
   fprintf(stderr, "Got %s clusters\n", temp);
   if (temp == NULL) {
      fprintf(logfile,
         "ERROR:  Numclusters must be defined as a positive integer\n");
      fprintf(stderr, "ERROR - see log file for details.\n");
      exit(1);
   }
   if (strcmp(temp, "CNUMCLUSTERS") == 0) {
      if (Test(5, temp) == 0)
         strcpy(temp, "1");
   }
   num_clusters = atoi(temp);
   if (num_clusters > numnodes)
      num_clusters = numnodes;
   cluster_size = numnodes / num_clusters;

   if ((num_clusters == numnodes) && (load_balancing != OFF)) {
      fprintf(logfile,
          "Warning:  Load balancing must be set to OFF when\n");
      fprintf(logfile,
          "   #clusters = #processors.  This has been done.\n");
      fprintf(stderr, "Warning:  Load balancing OFF - see log for details.\n");
      load_balancing = OFF;
   }

   strcpy(temp, getenv("optimal"));    /* optimal or first solution */
   if (temp == NULL ||
       ((strcmp(temp, "OPTIMAL") != 0) && (strcmp(temp, "SUBOPTIMAL") != 0))) {
      fprintf(logfile, "ERROR:  Optimal must = OPTIMAL or SUBOPTIMAL\n");
      fprintf(stderr, "ERROR:  see log file for details\n");
      exit(1);
   }
   if (!strcmp(temp, "OPTIMAL"))
      optimal = TRUE;
   else optimal = FALSE;

   strcpy(temp, getenv("anticipatory"));    /* OPEN list anticipatory value */
   if (temp == NULL) {
      fprintf(logfile,
         "ERROR:  Anticipatory must be defined as 0 or a positive integer\n");
      fprintf(stderr, "ERROR:  see log file for details.\n");
      exit(1);
   }
   if (strcmp(temp, "CANTICIP") == 0) {
      if (Test(6, temp) == 0)
         strcpy(temp, "1");
   }
   anticipatory = atoi(temp);
}


/* Set up the output log file. */
InitLog()
{
   strcpy(lfile, getenv("log"));
   if (lfile == NULL) {
      fprintf(stderr, "ERROR:  The environment var log must be set to a\n");
      fprintf(stderr, "        valid file for logging results\n");
      exit(1);
   }
   if ((logfile = fopen(lfile, "w")) == NULL) {
      fprintf(stderr, "Unable to open the logfile %s\n", logfile);
      exit(1);
   }
}


/* Output information about this run. */
OutputLog()
{
   fprintf(logfile, "Number of nodes    =   %d\n", numnodes);
   fprintf(logfile, "Number of clusters =   %d\n", num_clusters);
   fprintf(logfile, "Cluster size       =   %d\n", cluster_size);
   fprintf(logfile, "ALPHA              =   %f\n", alpha);

   if (distribution_type == KUMAR_RAO)
      fprintf(logfile, "Distribution type  =   Kumar & Rao\n");
   else if (distribution_type == DTS)
      fprintf(logfile, "Distribution type  =   Distributed Tree\n");

   switch (load_balancing) {
      case NEIGHBOR:
         fprintf(logfile, "Load balancing     =   Nearest Neighbor\n");
         break;
      case RANDOM:
         fprintf(logfile, "Load balancing     =   Random Selection\n");
         break;
      case ASYNC:
         fprintf(logfile, "Load balancing     =   Asynchronous Operation\n");
         break;
      default:
         fprintf(logfile, "Load balancing     =   OFF\n");
   }

   if (optimal == TRUE)
      fprintf(logfile, "Optimal/Suboptimal =   OPTIMAL\n");
   else fprintf(logfile, "Optimal/Suboptimal =   SUB-OPTIMAL\n");

   if (ordering == TOIDA)
      fprintf(logfile, "Ordering           =   TOIDA\n");
   else if (ordering == LOCAL)
      fprintf(logfile, "Ordering           =   LOCAL\n");
   else fprintf(logfile, "Ordering           =   NONE\n");

   if (giveaway == TOP)
      fprintf(logfile, "Giveaway           =   TOP\n");
   else fprintf(logfile, "Giveaway           =   BOTTOM\n");
   fprintf(logfile, "Anticipatory       =  %2d\n", anticipatory);
}


InitializeNodes()
{
   int i, numt;

                        /* Load node program on allocated processors */
   numt = pvm_spawn(SLAVE, (char **)0, 0, "", numnodes, tids);
   if (numt < numnodes) {
      fprintf(stderr, "Trouble spawning slave processors\n");
      for (i=0; i<numt; i++)
         pvm_kill(tids[i]);
      pvm_exit();
      exit(1);
   }
}


/* Generate nodes to distribute to processors. */
InitialDistribution(BFS, distribution_type, puzzle, base, cluster_size)
   List *BFS;
   int distribution_type, puzzle[], base, cluster_size;
{
   int i, j, dest, first, last, data[MAXOPS];
   Node node;

   /* if KUMAR_RAO, send root node to processor 0 */
   /* the normal node request message passes the path to the node */
   /* along with the actual node.  Since this is the root node,   */
   /* the path is empty and is omitted here                       */

   if (distribution_type == KUMAR_RAO) {
      pvm_initsend(PvmDataDefault);
      i = 1;
      pvm_pkint(&i, 1, 1);        /* number of nodes sent */
      i = -1;
      pvm_pkint(&i, 1, 1);        /* root node has no parent */
      i = 0;
      pvm_pkint(&i, 1, 1);        /* level of the root node */
      for (i=0; i<MAXOPS; i++)    /* none of the children have been generated */
         data[i] = 0;
      pvm_pkint(data, MAXOPS, 1);
      pvm_pkint(puzzle, 16, 1);
      pvm_pkint(&i, 1, 1);        /* path */
      pvm_pkint(&i, 1, 1);        /* number_children */
      i = state_cost(puzzle, goal);
      pvm_pkint(&i, 1, 1);        /* heuristic value */
      i = 0;
      pvm_pkint(&i, 1, 1);        /* cum_heuristic_error */
      pvm_send(tids[base], PUZZLE_COMING);
   }

   else {                             /* send distinct node to each processor */
      dest = base;
      first = BFS->First;
      last = BFS->Last;
      for (i=first; i>last; i--) {
         pvm_initsend(PvmDataDefault);
	 node = BFS->Elements[i];
         j = 1;
	 pvm_pkint(&j, 1, 1);
	 pvm_pkint(&node.parent, 1, 1);
	 pvm_pkint(&node.level, 1, 1);
	 pvm_pkint(node.child_expanded, MAXOPS, 1);
	 pvm_pkint(node.puzzle, 16, 1);
	 pvm_pkint(node.path, node.level+1, 1);
	 pvm_pkint(&node.number_children, 1, 1);
	 pvm_pkint(&node.h, 1, 1);
	 pvm_pkint(&node.cum_heuristic_error, 1, 1);
	 pvm_send(tids[dest], PUZZLE_COMING);
	 dest = ((dest + 1) % cluster_size) + base;
      }
   }

   for (i=0; i<cluster_size; i++) {
      pvm_initsend(PvmDataDefault);     /* send processors their cluster root */
      pvm_pkint(&base, 1, 1);
      pvm_send(tids[base+i], CLUSTER_ROOT);

      pvm_initsend(PvmDataDefault);     /* send processors their cluster size */
      pvm_pkint(&cluster_size, 1, 1);
      pvm_send(tids[base+i], CLUSTER_SIZE);
   }
}


int update_threshold (processor_status, cluster_num)
     int processor_status[], cluster_num;
{
   int i, j, lh, skew, temp[MAXOPS], idle_counter, sum;
   float percents[MAXOPS];


   idle_counter = 0;                  /* check to see if this cluster is idle */
   for (i=0; i<cluster_size; i++)
      idle_counter += processor_status[cluster_num * cluster_size + i];

         /* tell all processors in this cluster to stop the current iteration */
   if (idle_counter == cluster_size &&
       threshold_status[cluster_num][1] == FALSE) {
      for (i=0; i<cluster_size; i++) {
         pvm_initsend(PvmDataDefault);
         pvm_send(tids[cluster_num * cluster_size + i], HALT_PROCESSING);
      }

                  /* get lowest heuristic and new oporder from each processor */
      for (i=0; i<cluster_size; i++) {
         pvm_recv(-1, LOWEST_HEURISTIC_COMING);
         pvm_upkint(&j, 1, 1);
         pvm_upkint(temp, MAXOPS, 1);

         if (j < lowest_heuristic) {      /* a lower heuristic has been found */
            lowest_heuristic = j;
            for (j=0; j<MAXOPS; j++)
               oporder[j] = temp[j];
         }
      }
      for (i=0; i<cluster_size; i++) {
         pvm_recv(-1, STATS_COMING);
         pvm_upkint(&sum, 1, 1);
         pvm_upkint(&parents, 1, 1);
         pvm_upkint(node_count, MAXOPS, 1);
	 pvm_upkint(&lh, 1, 1);
         pvm_upkfloat(&avg_herror, 1, 1);
      }
      bf = (float) sum / (float) parents;         /* average branching factor */
      total_nodes += sum;                  /* total number of generated nodes */
      if (prev_nodes != 0)              /* average heuristic branching factor */
         hbf = (float) total_nodes / (float) prev_nodes;
      else hbf = 0.00;
      prev_nodes = total_nodes;

      sum = 0;
      for (i=0; i<MAXOPS; i++)
         sum += node_count[i];
      for (i=0; i<MAXOPS; i++)          /* proportion of tree under this node */
         percents[i] = (float) node_count[i] / sum;
      skew = sum / 2;
      j = 0;
      for (i=0; i<MAXOPS; i++) {
         j += node_count[i];
	 if (j >= skew) {
	    skew = i;
	    break;
         }
      }
      if (sum == 0)
	 imbalance = 0;
      else imbalance = stddev(MAXOPS, percents);
         fprintf(logfile,
	 "t %d b %5.2f hbf %5.2f herr %5.2f imb %5.2f lh %d skew %d nodes %d\n",
            threshold, bf, hbf, avg_herror, imbalance, lh, skew, total_nodes);
         fflush(logfile);

           /* wait for all processors in this cluster to send GOFORIT message */
      for (i=0; i<cluster_size; i++)
         pvm_recv(-1, GOFORIT);

      fprintf(stderr, "sample %d, total_nodes %d\n", sample, total_nodes);
      if ((sample == 1) && (total_nodes >= SAMPLESIZE))    /* sample finished */
         return(1);

      threshold += THRESHOLD_INCREMENT; /* send new threshold to this cluster */
      fprintf (logfile, "Sending threshold of %3d to cluster %3d\n",
               threshold, cluster_num);
      t2 = time((time_t *) 0);
      fprintf(stderr, "Sending threshold of %3d to cluster %3d at %s",
             threshold, cluster_num, ctime(&t2));  fflush(stderr);
      for (i=0; i<cluster_size; i++) {
         pvm_initsend(PvmDataDefault);
         pvm_pkint(&threshold, 1, 1);
         pvm_send(tids[cluster_num * cluster_size + i], THRESHOLD_COMING);
      }
      threshold_status[cluster_num][0] = threshold;  /* new cluster threshold */
      threshold_status[cluster_num][1] = FALSE;       /* cluster not finished */
      threshold_status[cluster_num][2] = -1;      /* no winning processor yet */

      pvm_initsend(PvmDataDefault);               /* send current xform order */
      pvm_pkint(oporder, MAXOPS, 1);
      pvm_mcast(tids, numnodes, OPORDER_COMING);

      pvm_initsend(PvmDataDefault);                  /* let processors resume */
      pvm_mcast(tids, numnodes, GOFORIT);
      for (i=0; i<cluster_size; i++)
         processor_status[cluster_num * cluster_size + i] = 0;
   }

   return(0);
}


/* Returns 1 if no outstanding thresholds less than smallest goal found.   */
int least_cost_goal()
{
   int i, least_cost;

   goal_cluster = -1;                      /* identify least-cost goal so far */
   least_cost = 999;
   for (i=0; i<num_clusters; i++)
      if ((threshold_status[i][1] == TRUE) &&           /* cluster found goal */
          (threshold_status[i][0] < least_cost)) {
         least_cost = threshold_status[i][0];  /* smallest finished threshold */
	 goal_cluster = i;
      }

             /* check for outstanding thresholds smaller than least-cost goal */
   if (goal_cluster != -1) {                 /* if any clusters have finished */
      if (optimal == FALSE)     /* want first solution, so this is least cost */
         return(1);
      for (i=0; i<num_clusters; i++)
         if (threshold_status[i][0] < least_cost)
	    return(0);
      return(1);
   }

   return(0);
}
