/************************************************************************

stats.c

Mean, Standard Deviation, Variance, and Student t Test statistics from
Numerical Recipes book.

MEAN(n,D) returns the mean of the n values in the D array.

STDDEV(n,D) returns the standard deviation of the n values in the D
array.

VARIANCE(n,D) returns the variance of the n values in the D array.

PAIRED_TTEST(n,D1,D2) returns the significance level (the lower the
better) that the difference between the n data points in D1 and D2 is
significant.  This is a one-sided, upper-tailed, matched-pairs test
where the null hypothesis is that the mean of D1 is greater than or equal
to the mean of D2.  The returned significance level is the probability
that the null hypothesis can be accepted, or one minus the probability
that the null hypothesis can be rejected and the mean of D1 is greater
than the mean of D2.

***********************************************************************/

#include "eureka.h"

#define ITMAX 100
#define EPS   3.0e-7


float D1[100], D2[100];

float mean(n,D)
   int n;
   float D[];
{
   int i;
   float sum = 0.0;

   for (i=0; i<n; i++) sum += D[i];
   return sum / ((float) n);
}


float variance(n,D)
   int n;
   float D[];
{
   int i;
   float m, sum = 0.0;

   m = mean(n,D);
   for (i=0; i<n; i++) sum += ((D[i] - m)*(D[i] - m));
   return sum / ((float) (n-1));
}


float stddev(n,D)
   int n;
   float D[];
{
   double s;

   s = sqrt((double) variance(n, D));
   return((float) s);
}
