/*************************************************************************/
/*								 	 */
/*	User interface for consulting trees and rulesets	 	 */
/*	------------------------------------------------	 	 */
/*								 	 */
/*************************************************************************/


#include "defns.i"


typedef	struct	ValRange	*RangeDescRec;
struct	ValRange
	{
	    Boolean		Known, Asked;
	    float		LowerBound, UpperBound, *Probability;
	};

extern	RangeDescRec		RangeDesc;


#define Fuzz 0.01


/*************************************************************************/
/*									 */
/*	Ask for the value of attribute Att if necessary			 */
/*									 */
/*************************************************************************/


    CheckValue(Att, Data)
/*  ----------  */
    Attribute Att;
    char Data[5][50];
{
    if ( RangeDesc[Att].Asked ) return;

    /*printf("%s:  ", AttName[Att]);*/
    ReadRange(Att, Data);
}


/*************************************************************************/
/*									 */
/*	Read a range of values for attribute Att or <cr>		 */
/*									 */
/*************************************************************************/


    ReadRange(Att, Data)
/*  ----------  */
    Attribute Att;
    char Data[5][50];
{
    RangeDesc[Att].Asked=true;

    if (strcmp(Data[Att], "?") == 0)
    {
       RangeDesc[Att].Known = false;
       return;
    }

    RangeDesc[Att].Known = true;

    if ( MaxAttVal[Att] )
	ReadDiscr(Att, Data);
    else ReadContin(Att, Data);
}



/*************************************************************************/
/*									 */
/*	Read a discrete attribute value or range			 */
/*									 */
/*************************************************************************/

    ReadDiscr(Att, Data)
/*  ---------  */
    Attribute Att;
    char Data[5][50];
{
    char Name[500];
    Boolean ReadName();
    DiscrValue dv, PNo;
    float P, PSum;

    ForEach(dv, 1, MaxAttVal[Att])
	RangeDesc[Att].Probability[dv] = 0.0;

    strcpy(Name, Data[Att]);
    printf("Read name %s for Att %d\n", Name, Att);

    dv = Which(Name, AttValName[Att], 1, MaxAttVal[Att]);
    P = 1.0;		/*  only one attribute value  */
    RangeDesc[Att].Probability[dv] = P;

    /*  Check that sum of probabilities is not > 1  */

    PNo = MaxAttVal[Att];
    PSum = 1.0;
    ForEach(dv, 1, MaxAttVal[Att])
	if ( RangeDesc[Att].Probability[dv] > Fuzz )
	    PSum -= RangeDesc[Att].Probability[dv];
	    PNo--;

    /*  Distribute the remaining probability equally among
	the unspecified attribute values  */

    PSum /= PNo;
    ForEach(dv, 1, MaxAttVal[Att])
	if ( RangeDesc[Att].Probability[dv] < Fuzz )
	    RangeDesc[Att].Probability[dv] = PSum;
}



/*************************************************************************/
/*									 */
/*	Read a continuous attribute value or range			 */
/*									 */
/*************************************************************************/


    ReadContin(Att, Data)
/*  ----------  */
    Attribute Att;
    char Data[10][10];
{
    sscanf(Data[Att], "%f", &RangeDesc[Att].LowerBound);
    RangeDesc[Att].UpperBound = RangeDesc[Att].LowerBound;
    printf("Lower %f Upper %f for Att %d\n",
           RangeDesc[Att].LowerBound, RangeDesc[Att].UpperBound, Att);
}
